# -*- coding: utf-8 -*-
"""
Created on Fri Sep 08 17:42:17 2017

@author: Yongguang Gong
"""

from __future__ import division,print_function
import calendar
import dateutil.parser
import numpy as np
import copy
import datetime
from functools import wraps
import time


def str2date(x):
    if isinstance(x,(str,np.str_,np.string_)):
        return dateutil.parser.parse(x)
    
    xtypes=isinstance(x,np.ndarray)
    if xtypes: x=x.tolist()
    
    y=copy.deepcopy(x)
    for i in range(len(x)):
        if isinstance(x[i],str):
            y[i]=dateutil.parser.parse(x[i])
            
        elif isinstance(x[i],list):
            for j in range(len(x[i])):
                y[i][j]=dateutil.parser.parse(x[i][j])
            
    if xtypes: y=np.array(y)
    return y



def date2str(x,format_flag):
    
    def __date2str(x,format_flag):
        if format_flag==1:
            return x.strftime('%Y-%m-%d')
        elif format_flag==0:
            return x.strftime('%Y-%m-%d %H:%M')
    
    if isinstance(x,datetime.datetime):
        return __date2str(x,format_flag)
    
    xtypes=isinstance(x,np.ndarray)
    if xtypes: x=x.tolist()
    
    y=copy.deepcopy(x)
    for i in range(len(x)):
        if isinstance(x[i],datetime.datetime):
            y[i]=__date2str(x[i],format_flag)
            
        elif isinstance(x[i],list):
            for j in range(len(x[i])):
                y[i][j]=__date2str(x[i][j],format_flag)
            
    if xtypes: y=np.array(y)
    return y



def date2num(x):
    date0=datetime.datetime(1900,3,1,0,0,0)
    
    def __date2num(x):
        tmp=x-date0
        return float(tmp.days)+tmp.seconds/24./3600.+61.
    
    if isinstance(x,datetime.datetime):
        return __date2num(x)
    
    xtypes=isinstance(x,np.ndarray)
    if xtypes: x=x.tolist()
    
    y=copy.deepcopy(x)
    for i in range(len(x)):
        if isinstance(x[i],datetime.datetime):
            y[i]=__date2num(x[i])
            
        elif isinstance(x[i],list):
            for j in range(len(x[i])):
                y[i][j]=__date2num(x[i][j])
            
    if xtypes: y=np.array(y)
    return y



def num2date(x):
    date0=datetime.datetime(1900,3,1,0,0,0)
    
    def __num2date(x):
        return date0+datetime.timedelta(x-61.)
    
    if isinstance(x,(int,float,np.int32,np.float64)):
        return __num2date(x)
    
    xtypes=isinstance(x,np.ndarray)
    if xtypes: x=x.tolist()
    
    y=copy.deepcopy(x)
    for i in range(len(x)):
        if isinstance(x[i],(int,float,np.int32,np.float64)):
            y[i]=__num2date(x[i])
            
        elif isinstance(x[i],list):
            for j in range(len(x[i])):
                y[i][j]=__num2date(x[i][j])
            
    if xtypes: y=np.array(y)
    return y
    



def addtenor(x,tenor,endmonth_flag=0):
    
    def __addmonth(x,tenor_num,endmonth_flag):
        x1=x+dateutil.relativedelta.relativedelta(months=tenor_num)
        if endmonth_flag==1:
            _,lastdaynum=calendar.monthrange(x.year,x.month)
            _,tmplastdaynum=calendar.monthrange(x1.year,x1.month)
            if (x.day==lastdaynum)&(x1.day!=tmplastdaynum) :
                x1=datetime.datetime(x1.year,x1.month,tmplastdaynum,x1.hour,x1.minute,x1.second)
        return x1
    
    if isinstance(x,(int,float,np.int32,np.float64)):
        x_date,x_num=num2date(x),copy.deepcopy(x)
    elif isinstance(x,datetime.datetime):
        x_date,x_num=copy.deepcopy(x),date2num(x)
    
    tenor_num,tenor_measure=tenor[0],tenor[1]
    if tenor_measure==1:
        y_num=x_num+tenor_num
        y_date=num2date(y_num)
    elif tenor_measure==2:
        y_num=x_num+tenor_num*7
        y_date=num2date(y_num)
    elif tenor_measure==3:
        y_date=__addmonth(x_date,int(tenor_num),endmonth_flag)
        y_num=date2num(y_date)
    elif tenor_measure==4:
        y_date=__addmonth(x_date,int(tenor_num*12),endmonth_flag)
        y_num=date2num(y_date)
    
    if isinstance(x,datetime.datetime):
        return y_date
    elif isinstance(x,(int,float,np.int32,np.float64)):
        return y_num
    


class bzdayadj_class(object):
    def __init__(self,Holiday,Weekend=1):
        self.Holiday=np.array(Holiday)
        self.Weekend=Weekend
    
    def to_wd(self,date0,methods):
        if (methods==0)|(methods=="NA"):
            return date0
        elif (methods==1)|(methods=="F"):
            return self.following(date0)
        elif (methods==-1)|(methods=="P"):
            return self.preceding(date0)
        elif (methods==2)|(methods=="MF"):
            return self.modifyfollowing(date0)
    
    def following(self,date0):
        while 1==1:
            tmp1=all(self.Holiday!=date0)
            date1=num2date(date0)
            tmp2=(date1.weekday()!=5)&(date1.weekday()!=6)
            if (tmp1&(self.Weekend==0))|(tmp1&(self.Weekend==1)&tmp2): 
                break
            date0=date0+1
        return date0
    
    def preceding(self,date0):
        while 1==1:
            tmp1=all(self.Holiday!=date0)
            date1=num2date(date0)
            tmp2=(date1.weekday()!=5)&(date1.weekday()!=6)
            if (tmp1&(self.Weekend==0))|(tmp1&(self.Weekend==1)&tmp2): 
                break
            date0=date0-1
        return date0
    
    def modifyfollowing(self,date0):
        date1=self.following(date0)
        tmp0,tmp1=num2date(date0),num2date(date1)
        if tmp0.month!=tmp1.month:
            date1=self.preceding(date0)
        return date1



def daycount(x1,x2,methods):
    
    def __daycount(date1,date2,methods):
        if (methods==1)|(methods=="Act365"):
            return (date2-date1)/365.
        elif (methods==2)|(methods=="Act360"):
            return (date2-date1)/360.
        elif (methods==3)|(methods=="30I360"):
            date1,date2=num2date(date1),num2date(date2)
            Y1=date1.year; M1=date1.month; D1=date1.day
            Y2=date2.year; M2=date2.month; D2=date2.day
            _,lastdaynum1=calendar.monthrange(Y1,M1)
            _,lastdaynum2=calendar.monthrange(Y2,M2)
            if (M1==2)&(D1==lastdaynum1)&(M2==2)&(D2==lastdaynum2): D2=30
            if (M1==2)&(D1==lastdaynum1): D1=30
            if (D2==31)&((D1==30)|(D1==31)):D2=30
            if D1==31:D1=30
            return ((Y2-Y1)*360+(M2-M1)*30+(D2-D1))/360.
    
    
    if isinstance(x1,(int,float,np.int32,np.float64)):
        return __daycount(x1,x2,methods)
    
    xtypes=isinstance(x1,np.ndarray)
    if xtypes: x1=x1.tolist(); x2=x2.tolist()
    
    y=copy.deepcopy(x1)
    for i in range(len(x1)):
        if isinstance(x1[i],(int,float)):
            y[i]=__daycount(x1[i],x2[i],methods)
            
        elif isinstance(x1[i],list):
            for j in range(len(x1[i])):
                y[i][j]=__daycount(x1[i][j],x2[i][j],methods)
            
    if xtypes: y=np.array(y)
    return y



class settleday_class(bzdayadj_class):
    def to_date(self,date0,days):
        signs=np.sign(days)
        if signs==0:
            methods="NA"
        elif signs==1:
            methods="F"
        elif signs==-1:
            methods="P"
        
        for i in range(abs(int(days))):
            date0=date0+signs
            date0=self.to_wd(date0,methods)
        return date0




def timefn(fn):
    @wraps(fn)
    def measure_time(*args, **kwargs):
        t1=time.time()
        result=fn(*args,**kwargs)
        t2=time.time()
        print ("@timefn:"+fn.__name__+" took "+str(round(t2-t1,4))+" seconds")
        return result
    return measure_time




if __name__ == "__main__":
    x1=['2012-7-5','2012-7-6']
    y1=str2date(x1)
    x2=[['2012-7-5'],['2012-7-6']]
    y2=str2date(x2)
    x3=np.array(x1)
    y3=str2date(x3)
    x4=np.array(x1,ndmin=2)
    y4=str2date(x4)
    x5=np.array(x2)
    y5=str2date(x5)
    x6='2012-7-5'
    y6=str2date(x6)
    x7=x3[0]
    y7=str2date(x7)    
    
    z1=date2str(y1,1)
    z2=date2str(y2,1)
    z3=date2str(y3,1)
    z4=date2str(y4,1)
    z5=date2str(y5,1)
    z6=date2str(y6,1)
    print(type(y4[0,0]))
    
    
    U1=date2num(y1)
    U2=date2num(y2)
    U3=date2num(y3)
    U4=date2num(y4)
    U5=date2num(y5)
    U6=date2num(y6)
    
    V1=num2date(U1)
    V2=num2date(U2)
    V3=num2date(U3)
    V4=num2date(U4)
    V5=num2date(U5)
    V6=num2date(U6)
    V7=num2date(U3[0])
    
    print(addtenor(V1[0],[3,1],1))
    
    Holiday=['2017-9-15','2017-9-18','2017-9-29']
    date0='2017-9-29'
    Holiday=date2num(str2date(Holiday))
    date0=date2num(str2date(date0))
    bzdayadj=bzdayadj_class(Holiday,0).to_wd
    
    print(num2date(bzdayadj(date0,"F")),num2date(bzdayadj(date0,"P")),num2date(bzdayadj(date0,"MF")))
    
    date1=date2num(str2date('2017-2-28'))
    date2=date2num(str2date('2017-3-31'))
    print(daycount(date1,date2,3))
    
    date3=date2num(str2date('2017-9-14'))
    settleday=settleday_class(Holiday,1).to_date
    print (num2date(settleday(date3,2)))
    


# -*- coding: utf-8 -*-
"""
Created on Thu Aug 24 13:10:57 2017

@author: Yongguang Gong
"""

from __future__ import division, print_function

import numpy as np
import scipy.stats as stats


class VanillaPricer_Class():
    
    def __init__(self,L,K,phi,s0,repo,div,r,vol,td,tv):
        self.L=L; self.K=K; self.phi=phi
        self.s0=s0; self.repo=repo; self.div=div
        self.r=r; self.vol=vol; self.td=td; self.tv=tv
        
    def AnalyticFormula(self):
        L=self.L; K=self.K; phi=self.phi
        s0=self.s0; repo=self.repo; div=self.div
        r=self.r; vol=self.vol; td=self.td; tv=self.tv
        
        forward=s0*np.exp((repo-div)*td)
        d1=np.log(forward/K)/vol/np.sqrt(tv)+0.5*vol*np.sqrt(tv)
        d2=d1-vol*np.sqrt(tv)
        OptPrice=phi*L*(forward*stats.norm.cdf(phi*d1)-K*stats.norm.cdf(phi*d2))*np.exp(-r*td)
        return OptPrice
        
        
    def MonteCarlo(self,PathNum,StepNum):
        L=self.L; K=self.K; phi=self.phi
        s0=self.s0; repo=self.repo; div=self.div
        r=self.r; vol=self.vol; td=self.td; tv=self.tv
        
        repo=repo*td/tv
        div=div*td/tv
        dt=tv/StepNum
        RandNum=np.random.normal(0.,1.,[StepNum,PathNum])
        RandNum=np.hstack((RandNum,-RandNum))
        SPath=s0*np.exp(np.cumsum((repo-div-0.5*vol**2)*dt+vol*RandNum*np.sqrt(dt),axis=0))
        OptPrice=L*np.mean(np.maximum(phi*(SPath[-1,:]-K),0))*np.exp(-r*td)
        return OptPrice
    
    
    def FiniteDifference(self,ns,nt):
        # by Crank-nicolson scheme
        L=self.L; K=self.K; phi=self.phi
        s0=self.s0; repo=self.repo; div=self.div
        r=self.r; vol=self.vol; td=self.td; tv=self.tv
        
        Smax=K*5.
        repo=repo*td/tv
        div=div*td/tv    
        r=r*td/tv
        
        s_grid=np.linspace(0,Smax,ns+1)
        dt=tv/nt
        i=np.array(range(ns+1))
        a1=0.25*(repo-div)*i*dt-0.25*vol**2*i**2*dt
        b1=1+0.5*vol**2*i**2*dt
        c1=-0.25*(repo-div)*i*dt-0.25*vol**2*i**2*dt
    
        a2=-0.25*(repo-div)*i*dt+0.25*vol**2*i**2*dt
        b2=1-0.5*vol**2*i**2*dt
        c2=0.25*(repo-div)*i*dt+0.25*vol**2*i**2*dt
    
        B1=np.diag(a1[2:-1],-1)+np.diag(b1[1:-1])+np.diag(c1[1:-2],1)
        B1_inverse=np.linalg.inv(B1)
        B2=np.diag(a2[2:-1],-1)+np.diag(b2[1:-1])+np.diag(c2[1:-2],1)
    
        BD_t=np.maximum(phi*(s_grid-K),0)
        j=np.array(range(nt+1))
        if phi==1:
            BD_Smin=np.zeros(nt+1)
            BD_Smax=np.maximum(Smax*np.exp((repo-div)*j*dt)-K,0)
        else:
            BD_Smax=np.zeros(nt+1)
            BD_Smin=K*np.ones(nt+1)
    
        u=BD_t
        for tj in range(1,nt+1):
            u=np.reshape(u,(len(u),1))
            g1=np.hstack((a1[1]*BD_Smin[tj],np.zeros(ns-3),c1[-2]*BD_Smax[tj]))
            g2=np.hstack((a2[1]*u[0],np.zeros(ns-3),c2[-2]*u[-1]))
            g1=np.reshape(g1,(len(g1),1))
            g2=np.reshape(g2,(len(g2),1))
        
            u[1:-1]=np.dot(B1_inverse,np.dot(B2,u[1:-1])+g2-g1)
            u[0]=BD_Smin[tj]
            u[-1]=BD_Smax[tj]
            u=u*np.exp(-r*dt)
            
        u=np.reshape(u,(len(u,)))    
        OptPrice=L*np.interp(s0,s_grid,u)
        return OptPrice



if __name__ == "__main__":
    L=10000.
    K=100. 
    phi=1.
    s0=105
    repo=0.04
    div=0.01
    r=0.03
    vol=0.5
    td=1.0
    tv=1.03
    PathNum=20000
    StepNum=100
    ns=100
    nt=80
    
    Pricer1=VanillaPricer_Class(L,K,phi,s0,repo,div,r,vol,td,tv)
    
    Price0=Pricer1.AnalyticFormula()
    Price1=Pricer1.MonteCarlo(PathNum,StepNum)
    Price2=Pricer1.FiniteDifference(ns,nt)
    
    print('Analytical Formula:',Price0)
    print('Monte Carlo:',Price1)
    print('Finite Difference:',Price2)
    


# -*- coding: utf-8 -*-
"""
Created on Thu Sep 14 13:37:39 2017

@author: Yongguang Gong
"""

from __future__ import division, print_function

from . import iDate
import numpy as np
import copy
from . import iMath


class discount_class(object):
    def __init__(self,Curve,SpotDate,CurveSettings):
        if CurveSettings[0]=="T":
            if CurveSettings[1]=="Act365":
                Curve[:,0]=(Curve[:,0]-SpotDate)/365.
            elif CurveSettings[1]=="Act360":
                Curve[:,0]=(Curve[:,0]-SpotDate)/360.
        self.Curve=Curve
        self.SpotDate=SpotDate
        self.CurveSettings=CurveSettings
    
    def to_dscnt(self,Date0):
        if self.CurveSettings[1]=="Act365":
            t0=(Date0-self.SpotDate)/365.
        elif self.CurveSettings[1]=="Act360":
            t0=(Date0-self.SpotDate)/360.
        
        rate0=iMath.interp1(self.Curve[:,0],self.Curve[:,1],t0,'linear')
        if self.CurveSettings[2]=="C":
            return np.exp(-rate0*t0)


class forwardrate_class(discount_class):
    def __init__(self,Curve,SpotDate,CurveSettings,IndexSettings):
        super(forwardrate_class,self).__init__(Curve,SpotDate,CurveSettings)
        self.IndexSettings=IndexSettings
        
    def to_fwd(self,effect,expiry):
        dscnt1=self.to_dscnt(effect)
        dscnt2=self.to_dscnt(expiry)
        
        if self.IndexSettings[0]=="Act365":
            return (dscnt1/dscnt2-1)/(expiry-effect)*365.
        elif self.IndexSettings[0]=="Act360":
            return (dscnt1/dscnt2-1)/(expiry-effect)*360.



class IRSPricer_class(object):
    
    def __init__(self,SpotDate,RefRateFlag,RefCurve,RefRate_his,DisCurve,Holiday):
        self.SpotDate=SpotDate[0]
        self.Tenor=RefRateFlag
        self.RefCurve=np.array(RefCurve)
        self.RefRate_his=np.array(RefRate_his)
        self.DisCurve=np.array(DisCurve)
        self.bzdayadj=iDate.bzdayadj_class(Holiday,1).to_wd
        self.settleday=iDate.settleday_class(Holiday,1).to_date
        self.Acc2PayFlag=1
        self.fixdatesettledays=-2
        self.IndexSettledays=2
        self.EndMonthFlag=1
        self.discount=discount_class(self.DisCurve,self.SpotDate,["D","Act365","C"]).to_dscnt
        self.forwardrate=forwardrate_class(self.RefCurve,self.SpotDate,["D","Act365","C"],["Act365"]).to_fwd
        
    
    def CFDate(self,EffectDate,ExpiryDate,Freq):
        Period=12./Freq
        N=int(np.round((ExpiryDate-EffectDate)/365.*Freq))
        CF_expiry,CF_paydate =np.zeros(N),np.zeros(N)
     
        for i in range(N):
            CF_expiry[i]=iDate.addtenor(EffectDate,[int(Period*(i+1)),3],self.EndMonthFlag)
            CF_paydate[i]=self.bzdayadj(CF_expiry[i],"F")
        CF_expiry[N-1]=ExpiryDate
        if self.Acc2PayFlag==1:
            CF_expiry=copy.deepcopy(CF_paydate)
        
        CF_effect=np.zeros(N); CF_effect[0]=EffectDate
        CF_effect[1:N]=copy.deepcopy(CF_expiry[0:N-1])
        
        return CF_effect,CF_expiry,CF_paydate
    
    
    
    def RollCFAmount(self,EffectDate,ExpiryDate,Tenor,Spread):
        
        Tenor_num,Tenor_measure=Tenor[0],Tenor[1]
        
        Rolling_expiry=iDate.addtenor(EffectDate,[Tenor_num,Tenor_measure],self.EndMonthFlag)
        Rolling_expiry=np.array([self.bzdayadj(Rolling_expiry,"F")])
        
        while Rolling_expiry[-1]<ExpiryDate:
            expiry=iDate.addtenor(Rolling_expiry[-1],[Tenor_num,Tenor_measure],self.EndMonthFlag)
            Rolling_expiry=np.hstack((Rolling_expiry,np.array([expiry])))
        
        Rolling_expiry=Rolling_expiry[Rolling_expiry<ExpiryDate]
        Rolling_expiry=np.hstack((Rolling_expiry,np.array([ExpiryDate])))
        Rolling_effect=np.hstack((np.array([EffectDate]),Rolling_expiry[0:-1]))
        
        N=len(Rolling_effect)
        Rolling_fixdate,Fwd_effect,Fwd_expiry=np.zeros(N),np.zeros(N),np.zeros(N)
        
        for i in range(N):
            Rolling_fixdate[i]=self.settleday(Rolling_effect[i],self.fixdatesettledays)
            Fwd_effect[i]=self.settleday(Rolling_fixdate[i],self.IndexSettledays)
            Fwd_expiry[i]=iDate.addtenor(Fwd_effect[i],Tenor,self.EndMonthFlag)
            Fwd_expiry[i]=self.bzdayadj(Fwd_expiry[i],"NA")
            
        RefRate=iMath.interp1(self.RefRate_his[:,0],self.RefRate_his[:,1],Rolling_fixdate,'linear')
        AccInt=np.zeros(N)
        for i in range(N):
            if Rolling_fixdate[i]>SpotDate:
                RefRate[i]=self.forwardrate(Fwd_effect[i],Fwd_expiry[i])
            AccInt[i]=(RefRate[i]+Spread)*iDate.daycount(Rolling_effect[i],Rolling_expiry[i],"Act360")
        CumAccInt=np.prod(1.+AccInt)-1
        return CumAccInt
    
    
    
    def FixLeg(self,Notional,EffectDate,ExpiryDate,Strike,Freq):
        EffectDate,ExpiryDate,Freq=EffectDate[0],ExpiryDate[0],Freq[0]
        Notional,Strike=Notional[0],Strike[0]
        
        CF_effect,CF_expiry,CF_paydate=self.CFDate(EffectDate,ExpiryDate,Freq)
        
        Term0=iDate.daycount(CF_effect,CF_expiry,"Act365")
        CF_amount=Notional*Strike*Term0
        
        CashFlow=np.transpose(np.vstack([CF_paydate,CF_amount]))
        
        CF_paydate=CF_paydate[CF_paydate>=self.SpotDate]
        CF_amount=CF_amount[CF_paydate>=self.SpotDate]
        
        if len(CF_paydate)==0: return CashFlow, 0.
        
        Dscnt=self.discount(CF_paydate)
        
        return CashFlow, np.sum(CF_amount*Dscnt)
        
        

    def FltLeg(self,Notional,EffectDate,ExpiryDate,Freq,Spread):
        EffectDate,ExpiryDate,Freq=EffectDate[0],ExpiryDate[0],Freq[0]
        Notional,Spread=Notional[0],Spread[0]
        
        CF_effect,CF_expiry,CF_paydate=self.CFDate(EffectDate,ExpiryDate,Freq)
        N=len(CF_effect)
        CF_amount=np.zeros(N)
        for i in range(N):
            CF_amount[i]=Notional*self.RollCFAmount(CF_effect[i],CF_expiry[i],[7,1],Spread)
        
        CashFlow=np.transpose(np.vstack([CF_paydate,CF_amount]))
        
        CF_paydate=CF_paydate[CF_paydate>=self.SpotDate]
        CF_amount=CF_amount[CF_paydate>=self.SpotDate]
        
        if len(CF_paydate)==0: return CashFlow, 0.
        
        Dscnt=self.discount(CF_paydate)
        
        return CashFlow, np.sum(CF_amount*Dscnt)
    
    
    
    def Swap(self,BuyFlag,Notional,EffectDate,ExpiryDate,Strike,Spread,Freq_fix,Freq_flt):
        BuyFlag=BuyFlag[0]
        CF_fix,Value_fix=self.FixLeg(Notional,EffectDate,ExpiryDate,Strike,Freq_fix)
        CF_flt,Value_flt=self.FltLeg(Notional,EffectDate,ExpiryDate,Freq_flt,Spread)
        
        return BuyFlag*(Value_fix-Value_flt)
        
 



if __name__ == "__main__":
    BuyFlag=[2]
    Notional=[100000]
    EffectDate='12-Aug-2007'
    ExpiryDate='12-Aug-2009'
    SpotDate='12-Aug-2007'
    Strike=[0.04]
    RefRate=[1]
    Freq_fix=[4]
    Freq_flt=[4]
    Spread=[0.005]
    DayCountFlag=[1]
    
    RefCurve=[[3/365.,0.0259],[7/365.,0.0328],[14/365.,0.0328],[1/12.,0.0338],[0.25,0.0432],[0.5,0.0457],[0.75,0.0464],[1.,0.0471],[2.,0.0477],[3.,0.0480],[4.,0.0486],[5.,0.0498],[6.,0.0510],[7.,0.0518],[8.,0.0522],[9.,0.0525],[10.,0.053]]
    DisCurve=[[3/365.,0.0259],[7/365.,0.0328],[14/365.,0.0328],[1/12.,0.0338],[0.25,0.0432],[0.5,0.0457],[0.75,0.0464],[1.,0.0471],[2.,0.0477],[3.,0.0480],[4.,0.0486],[5.,0.0498],[6.,0.0510],[7.,0.0518],[8.,0.0522],[9.,0.0525],[10.,0.053]]
    
    RefRate_his=list([[39927,0.0121]])
    for i in range(700):
        RefRate_his.append([39927-i,0.0121])
    
    
    EffectDate=[iDate.date2num(iDate.str2date(EffectDate))]
    ExpiryDate=[iDate.date2num(iDate.str2date(ExpiryDate))]
    SpotDate=[iDate.date2num(iDate.str2date(SpotDate))]
    
    IRSPricer1=IRSPricer_class(SpotDate,RefRate,RefCurve,RefRate_his,DisCurve,[1])
    CashFlow1,Value1=IRSPricer1.FixLeg(Notional,EffectDate,ExpiryDate,Strike,Freq_fix)
    
    CashFlow2,Value2=IRSPricer1.FltLeg(Notional,EffectDate,ExpiryDate,Freq_flt,Spread)
    


# -*- coding: utf-8 -*-
"""
Created on Sun Aug 30 09:34:10 2015

@author: Yongguang Gong
"""
from __future__ import division, print_function
import numpy as np
from scipy.stats import norm, lognorm
from scipy.integrate import quad
from . import iMath


class AsianPricer_class(object):
    def __init__(self,L,K,phi,s0,repo,div,vol,T):
        self.L = L; self.K = K; self.phi = phi
        self.s0 = s0; self.repo = repo; self.div= div
        self.vol = vol; self.T = T
    
    def GeometricAverage(self):
        L = self.L; s0=self.s0; r = self.repo
        q = self.div; K = self.K; Sigma = self.vol
        T = self.T; phi = self.phi
        
        SigmaA=Sigma/np.sqrt(3.0)
        ba=0.5*(r-q-Sigma*Sigma/6.0)
        d1=(np.log(s0/K)+(ba+0.5*SigmaA*SigmaA)*T)/SigmaA/np.sqrt(T)
        d2=d1-SigmaA*np.sqrt(T)
        OptVal=phi*L*(s0*np.exp((ba-r)*T)*norm.cdf(phi*d1,0.0,1.0)-K*np.exp(-r*T)*norm.cdf(phi*d2,0.0,1.0))
        
        return OptVal
    
    
    def CurranApprox(self):
        L = self.L; s0=self.s0; r = self.repo
        q = self.div; K = self.K; Sigma = self.vol
        T = self.T; phi = self.phi
        #dt=T/N
        dt = 1/252; N = round(T/dt)
        
        iy=np.array(range(1,N+1,1),ndmin=2)
        ix=np.transpose(iy)
        Mu_lnS=np.log(s0)+(r-q-0.5*Sigma**2.)*ix*dt
        Var_lnS=Sigma**2*ix*dt
        Corr_lnS=np.sqrt(iMath.Min(ix*np.ones((1,N)),iy*np.ones((N,1)))/iMath.Max(ix*np.ones((1,N)),iy*np.ones((N,1))))
        CoVar_lnS=np.sqrt(np.dot(Var_lnS,np.transpose(Var_lnS)))*Corr_lnS
        Mu_lnG=np.mean(Mu_lnS)
        Var_lnG=np.dot(np.sqrt(np.transpose(Var_lnS)),np.dot(Corr_lnS,np.sqrt(Var_lnS)))/N**2.
        Var_lnSG=np.sqrt(Var_lnS)*np.dot(Corr_lnS,np.sqrt(Var_lnS))/N
        
        Part21=np.exp(Mu_lnS+0.5*Var_lnS)*norm.cdf((Mu_lnG-np.log(K))/np.sqrt(Var_lnG)+Var_lnSG/np.sqrt(Var_lnG),0.,1.)
        Part21=np.mean(Part21)
        Part22=K*norm.cdf((Mu_lnG-np.log(K))/np.sqrt(Var_lnG))
        Part2=Part21-Part22
    
        CoVar_lnSonlnG=CoVar_lnS-np.dot(np.dot(CoVar_lnS,np.ones((N,1))),np.transpose(np.dot(CoVar_lnS,np.ones((N,1)))))/np.sum(np.sum(CoVar_lnS))
        Mu_lnSonlnG=Mu_lnS+Var_lnSG/Var_lnG*(np.log(K)-Mu_lnG)
        Mu_AonlnG=np.mean(np.exp(Mu_lnSonlnG+0.5*np.reshape(np.diag(CoVar_lnSonlnG),(-1,1))))
        Var_SonlnG=np.exp(Mu_lnSonlnG+0.5*np.reshape(np.diag(CoVar_lnSonlnG),(-1,1)))
        CoVar_SonlnG=(Var_SonlnG*np.ones((1,N)))*np.transpose(Var_SonlnG*np.ones((1,N)))*(np.exp(CoVar_lnSonlnG)-1.0)
        Var_AonlnG=np.sum(np.sum(CoVar_SonlnG))/N**2.
    
        Mu_Diff=Mu_AonlnG-K
        Var_Diff=np.copy(Var_AonlnG)
        Var_lnDiff=np.log(Var_Diff/Mu_Diff**2.+1.)
        Diff0=np.copy(Mu_Diff)
        es=0.000000000001
        K=K-es
    
        func=lambda x:iBSOption(Diff0,Var_lnDiff,K,x,Mu_lnG,Var_lnG)
        Part1,IntError=quad(func,es,K)
        callV = float(L*(Part1+Part2)*np.exp(-r*T))
        putV = callV-L*(s0*np.exp(-q*T)-K* np.exp(-r*T))
        
        return callV if phi==1 else putV


def iBSOption(Diff0,Var_lnDiff,K,G,Mu_lnG,Var_lnG):
    d1=(np.log(Diff0/(K-G)))/np.sqrt(Var_lnDiff)+0.5*np.sqrt(Var_lnDiff)
    d2=d1-np.sqrt(Var_lnDiff)
    OptVal=Diff0*norm.cdf(d1,0.,1.)-(K-G)*norm.cdf(d2,0.,1.)
    return OptVal*lognorm.pdf(G,np.sqrt(Var_lnG),0,np.exp(Mu_lnG))



if __name__ == "__main__":
    L=1.0
    s0=200.0
    r=0.04
    q=0.02
    K=200.0
    Sigma=0.4
    T=1.0
    phi=1.0
    
    import time
    time0=time.time()
    Call=AsianPricer_class(L,K,phi,s0,r,q,Sigma,T).GeometricAverage()
    print(Call, time.time()-time0)
    


# -*- coding: utf-8 -*-
"""
Created on Mon Jan 25 14:52:02 2016

@author: Yongguang Gong
"""

from __future__ import division,print_function
from scipy.stats import norm
from scipy.interpolate import interp1d
import ast
import numpy as np


def interp1(x,y,xi,interp_method):
    if (interp_method==1)|(interp_method=='linear'):
        func=interp1d(x,y,kind='linear',bounds_error=False,fill_value=(y[0],y[-1]))
    return func(xi)


def normcdf(x):
    return norm.cdf(x,0.0,1.0)


def str2var(x):
    return ast.literal_eval(x)


def col2row(x_columns):
    x_rows=list(range(len(x_columns)))
    for i in range(len(x_columns)):
        x_rows[i]=x_columns[i][0]
    return x_rows
    
    
def Max(a,b):
    c=np.copy(a)
    if np.size(b)==1:
        c[a<b]=b
    else:
        c[a<b]=b[a<b]
    return c


def Min(a,b):
    c=np.copy(a)
    if np.size(b)==1:
        c[a>b]=b
    else:
        c[a>b]=b[a>b]
    return c



if __name__ == "__main__":
    x=[1.,2.,3.,4.,5.,6.]
    y=[0.1,0.2,0.3,0.4,0.5,0.6]
    xi=[0.5,2.5,7.5]
    yi=interp1(x,y,xi,1)
    
    print(normcdf(x))
    print(str2var('2'))
    


# -*- coding: utf-8 -*-
"""
Created on Thu Jul  8 10:28:06 2021

@author: gyg
"""

import json


def save_json(dicts, filepath):
    with open(filepath, 'w') as result_file:
        json.dump(dicts, result_file)
        
        return True

def read_json(filepath):
    with open(filepath, 'r') as result_file:
        return json.load(result_file)


# -*- coding: utf-8 -*-
"""
Created on Tue Jul  6 18:44:20 2021

@author: Yongguang Gong
"""

'''
tutorial
0. python setup.py check
1. python setup.py sdist bdist_wheel
2. twine upload dist/*
load: juliangong/gyg3380149
3. https://pypi.org/pypi?%3Aaction=list_classifiers
4. python setup.py develop
'''

import setuptools

with open("README.md",'r') as fh:
    long_description = fh.read()
    
setuptools.setup(
    name = "ehyena",
    version = "0.0.1",
    author = "Julian Gong",
    author_email = "juliangong@hotmail.com",
    description = "option pricer",
    long_description = long_description,
    long_description_content_type = "text/markdown",
    url = "https://www.optprc.com",
    # install_requires= ["numpy>=1.22", ],
    packages = setuptools.find_packages(),
    classifiers = [
            #"Topic::Financial",
            #"Topic::Software Development::Libraries::Python Modules",
            "Programming Language :: Python :: 2.7",
            "Programming Language :: Python :: 3",
            "License :: OSI Approved :: MIT License",
            "Operating System :: OS Independent",
    ],
     
    )




# -*- coding: utf-8 -*-
"""
Created on Thu Nov 09 13:09:10 2017

@author: Yongguang Gong
"""

from __future__ import absolute_import

__name__ = "option pricer"
__version__ = "0.1.2"

from .IRS import IRSPricer_class as IRSPricer
from .Vanilla import VanillaPricer_Class as VanillaPricer
from .Asian import AsianPricer_class as AsianPricer
from .iData import save_json, read_json
from .iDate import timefn

__all__ = ["IRSPricer","VanillaPricer","AsianPricer","timefn","save_json","read_json"]